#!/usr/bin/env python
# Copyright 2018 ARC Centre of Excellence for Climate Extremes
# author: Scott Wales <scott.wales@unimelb.edu.au>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Refactored  by Jost von Hardenberg <jost.hardenberg@polito.it>
# and Paolo Davini <paolo.davini@cnr.it>

"""Dask-aware regridding

To apply a regridding you will need a set of weights mapping from the source
grid to the target grid.

Regridding weights can be generated online using CDO (:class:`CdoGenerate`),
or offline by calling these programs externally (this is recommended especially
for large grids).

Once calculated :func:`regrid` will apply these weights using a Dask sparse
matrix multiply, maintaining chunking in dimensions other than lat and lon.

:class:`Regrid` can create basic weights and store them to apply the weights to
multiple datasets.
"""


import math
import os
import xarray
import numpy
import dask.array
from .dimension import remove_degenerate_axes
from .cdogenerate import CdoGenerate
from .weights import compute_weights_matrix3d, compute_weights_matrix, mask_weights, check_mask
from .log import setup_logger
from .gridinspector import GridInspector
from .util import deprecated_argument, detect_nan_variation_dims, tolist

DEFAULT_AREA_MIN = 0.5  # default minimum area for conservative remapping


class Regridder(object):
    """Main smmregrid regridding class"""

    def __init__(self, source_grid=None, target_grid=None, weights=None,
                 method='con', remap_area_min=DEFAULT_AREA_MIN, transpose=True, mask_dim=None, vertical_dim=None,
                 horizontal_dims=None, cdo_extra=None, cdo_options=None,
                 check_nan=False,
                 cdo='cdo', loglevel='WARNING'):
        """
        Initialize the Regridder for performing regridding operations.

        This class allows for regridding between two grids, either by providing
        source and target grids or by using pre-computed weights. For large grids,
        it is recommended to pre-calculate weights using CDO. If weights are not
        supplied, they will be calculated using CDO's `genbil` function based on
        the provided source and target grids.

        Pre-computed weights can be generated externally or by using
        :class:`CdoGenerate()`.

        Args:
            source_grid (xarray.DataArray or str): Source grid dataset or file path.
            target_grid (xarray.DataArray): Target grid dataset for regridding.
            weights (xarray.Dataset): Pre-computed interpolation weights.
            mask_dim (str): Name of the masked coordinate for
                                3D weights generation (default: None).
            horizontal_dims (list): List of spatial dimensions to
                                    interpolate (default: None).
            method (str): Interpolation method to use (default: 'con').
            remap_area_min (float): Minimum area for remapping in conservative remapping.
                                    Larger values avoid mask erosion.
            transpose (bool): If True, transpose the output such that
                              the masked coordinate is placed just before the
                              other spatial coordinates (default: True).
            check_nan (bool): If True, check for NaN variation in the source grid
                              to define the masked dimension.
            cdo (str): Path to the CDO executable (default: 'cdo').
            cdo_extra (str): Extra command to be passed to CDO
            cdo_options(str): Extra options to be passed to CDO
            loglevel (str): Logging level for the operation (default: 'WARNING').

        Raises:
            ValueError: If neither weights nor source and target grids are provided.
            FileNotFoundError: If the specified source grid file does not exist.
            KeyError: If no grid types are found in the data.

        Warnings:
            DeprecationWarning: If deprecated arguments 'vertical_dim` is used. 
        """

        if (source_grid is None or target_grid is None) and (weights is None):
            raise ValueError(
                "Either weights or source_grid/target_grid must be supplied"
            )

        mask_dim = deprecated_argument(vertical_dim, mask_dim, 'vertical_dim', 'mask_dim')

        # set up logger
        self.loggy = setup_logger(level=loglevel, name='smmregrid.Regrid')
        self.loglevel = loglevel
        self.transpose = transpose
        mask_dim = tolist(mask_dim)
        horizontal_dims = tolist(horizontal_dims)
        if mask_dim:
            self.loggy.info('Forcing mask_dim from input: expecting a single-gridtype dataset')
        if horizontal_dims:
            self.loggy.info('Forcing horizontal_dim from input: expecting a single-gridtype dataset')
        self.extra_dims = {'mask': mask_dim, 'horizontal': horizontal_dims}

        # TODO: this might be overridden at regrid level
        self.loggy.debug("Minimum remap area: %s", remap_area_min)
        self.remap_area_min = float(remap_area_min)
        if self.remap_area_min < 0.0 or self.remap_area_min > 1.0:
            raise ValueError('The remap_area_min provided must be between 0.0 and 1.0')
        # Is there already a weights file?
        if weights is not None:
            self.loggy.info('Init from weights selected!')
            self.init_mode = 'weights'
            self.grids = self._gridtype_from_weights(weights)
        else:
            self.loggy.info('Init from grids selected!')
            self.init_mode = 'grids'
            if isinstance(source_grid, str):
                if os.path.isfile(source_grid):
                    source_grid_array = xarray.open_dataset(source_grid)
                else:
                    raise FileNotFoundError(f'Cannot find grid file {source_grid}')
            else:
                source_grid_array = source_grid

            self.grids = self._gridtype_from_data(source_grid_array)

            len_grids = len(self.grids)
            if len_grids == 0:
                raise ValueError('Cannot find any gridtype in your data, aborting!')
            if len_grids == 1:
                self.loggy.info('One gridtype found! Standard procedure')
            else:
                self.loggy.warning('%s gridtypes found! We are in uncharted territory!', len_grids)

            for index, gridtype in enumerate(self.grids):
                if check_nan:
                    # check for NaN variation in the source grid
                    gridtype = self._check_nan_variation(source_grid_array, gridtype)
                    self.grids[index] = gridtype  # update reference

                self.loggy.info('Processing grid number %s', index)
                self.loggy.debug('Processing grids %s', gridtype.dims)
                self.loggy.debug('Horizontal dimensions are %s', gridtype.horizontal_dims)
                self.loggy.debug('Mask dimension is %s', gridtype.mask_dim)
                self.loggy.debug('Other dimensions are %s', gridtype.other_dims)
                self.loggy.debug('Variables are %s', gridtype.variables)

                # always prefer to pass filename (i.e. source_grid) when possible to CdoGenerate()
                # this will limit errors from xarray and speed up CDO itself
                # it wil work only for single-gridtype dataset
                if isinstance(source_grid, str) and len_grids == 1:
                    source_grid_array_to_cdo = source_grid
                else:
                    # when feeding from xarray, select the variable and its bounds
                    if isinstance(source_grid_array, xarray.Dataset):
                        stored_vars = [list(gridtype.variables.keys())[0]] + gridtype.bounds
                        self.loggy.debug('Storing variables %s', stored_vars)
                        source_grid_array_to_cdo = source_grid_array[stored_vars]
                    else:
                        source_grid_array_to_cdo = source_grid_array

                    if gridtype.time_dims:
                        self.loggy.debug('Selecting only first time step for dimension %s', gridtype.time_dims[0])
                        source_grid_array_to_cdo = source_grid_array_to_cdo.isel({gridtype.time_dims[0]: 0})

                generator = CdoGenerate(source_grid_array_to_cdo, target_grid,
                                        cdo=cdo, cdo_options=cdo_options,
                                        cdo_extra=cdo_extra, loglevel=loglevel)
                gridtype.weights = generator.weights(method=method,
                                                     mask_dim=gridtype.mask_dim)

        for gridtype in self.grids:
            if gridtype.mask_dim:
                gridtype.weights_matrix = compute_weights_matrix3d(gridtype.weights,
                                                                   gridtype.mask_dim)
            else:
                gridtype.weights_matrix = compute_weights_matrix(gridtype.weights)

            # this section is used to create a target mask initializing the CDO weights (both 2d and 3d)
            # has a destination mask been precomputed?
            if "dst_grid_masked" in gridtype.weights.variables:
                gridtype.masked = gridtype.weights.dst_grid_masked.data
            else:
                # compute the destination mask now
                gridtype.weights = mask_weights(gridtype.weights, gridtype.weights_matrix, gridtype.mask_dim)
                gridtype.masked = check_mask(gridtype.weights, gridtype.mask_dim)

    def _gridtype_from_weights(self, weights):
        """
        Initialize the gridtype reading from weights
        """

        self.loggy.info('Precomputed weights support so far single-gridtype datasets')

        if not isinstance(weights, xarray.Dataset):
            weights = xarray.open_mfdataset(weights)

        grid_info = GridInspector(weights, cdo_weights=True, extra_dims=self.extra_dims,
                                  clean=False, loglevel=self.loglevel)
        gridtype = grid_info.get_gridtype()

        # the vertical dimension has to show up into the extra dimensions
        # to cover the case that it is not a standard dimension: possibly better implementation available
        self.extra_dims['mask'] = [gridtype[0].mask_dim]

        return gridtype

    def _gridtype_from_data(self, source_grid_array):
        """
        Initialize the gridtype reading from source_data
        """
        grid_info = GridInspector(source_grid_array, extra_dims=self.extra_dims,
                                  clean=True, loglevel=self.loglevel)
        return grid_info.get_gridtype()

    def regrid(self, source_data):
        """
        Regrid the provided source data to match the target grid.

        This method applies the regridding process to either a single data array or
        a dataset containing multiple data variables.

        Args:
            source_data (xarray.DataArray or xarray.Dataset): The source data to be regridded.

        Returns:
            xarray.DataArray or xarray.Dataset: The regridded data, matching the target grid.

        Raises:
            TypeError: If the provided source data is neither an xarray DataArray nor Dataset.
        """

        # apply the regridder on each DataArray
        if isinstance(source_data, xarray.Dataset):

            # extra call to GridInspector for the raise with multiple grids
            grid_inspect = GridInspector(source_data,
                                         extra_dims=self.extra_dims,
                                         loglevel=self.loglevel)
            datagrids = grid_inspect.get_gridtype()
            if len(datagrids) > 1 and self.init_mode == 'weights':
                raise ValueError(f'Cannot process data with {len(datagrids)} GridType initializing from weights')

            # map on multiple dataarray
            out = source_data.map(self.regrid_array, keep_attrs=True)

            # clean from degenerated variables
            degen_vars = [var for var in out.data_vars if out[var].dims == ()]
            return out.drop_vars(degen_vars)

        if isinstance(source_data, xarray.DataArray):
            return self.regrid_array(source_data)

        raise TypeError('The object provided is not a Xarray object!')

    def regrid_array(self, source_data):
        """
        Perform regridding on a single 2D or 3D array.

        This method decides whether to call the 2D or 3D regridding function based
        on the dimensions of the source data.

        Args:
            source_data (xarray.DataArray): The source data array to regrid.

        Returns:
            xarray.DataArray: The regridded array.

        Raises:
            ValueError: If the input data does not match expected dimensions.
        """
        scalars = self._identify_scalar_coords(source_data)
        if scalars:
            self.loggy.warning(
                "Found scalar coordinates %s. If have selected a along a masked dimensions,"
                "regridding might fail. "
                "Please consider subsetting with [] or with slice", scalars)

        self.loggy.debug('Getting GridType from source_data')
        grid_inspect = GridInspector(source_data,
                                     extra_dims=self.extra_dims,
                                     loglevel=self.loglevel)
        datagrids = grid_inspect.get_gridtype()

        for datagridtype in datagrids:
            if datagridtype.mask_dim:
                # if this is a 3D we specified the masked coord and it has it
                return self.regrid3d(source_data, datagridtype)
            # 2d case
            return self.regrid2d(source_data, datagridtype)
        
        # fall back for the case that we have no gridtype,
        # or we have only scalar coordinates, 
        # or we are trying to regrid bounds variables
        return xarray.DataArray(data=None)

    def _identify_scalar_coords(self, source_data):
        """
        This method checks for coordinates that have no dimensions (i.e., scalar) and returns a list of their names.

        Args:
            source_data (xarray.DataArray): The source data array to check for scalar coordinates.
        """

        return [name for name, coord in source_data.coords.items() if coord.dims == ()]

    def _get_gridtype(self, datagridtype):

        # special case for CDO weights without any dimensional information
        # we derived this from the regridded data and we use it as it is
        if self.init_mode == 'weights':
            self.loggy.info('Assuming gridtype from data to be the same from weights')
            self.grids[0].dims = datagridtype.dims
            self.grids[0].horizontal_dims = datagridtype.horizontal_dims
            self.grids[0].other_dims = datagridtype.other_dims

        # match the grid
        gridtype = next((grid for grid in self.grids if grid == datagridtype), None)

        return gridtype

    def regrid3d(self, source_data, datagridtype):
        """
        Regrid a 3D source data array to match the target grid.

        This method applies the necessary weights and handles masked coordinates.

        Args:
            source_data (xarray.DataArray): The 3D source data to be regridded.
            datagridtype: The grid type information for the source data.

        Returns:
            xarray.DataArray: The regridded 3D data.

        Raises:
            ValueError: If there are issues with transposing the output dimensions.
        """

        self.loggy.debug('3D DataArray access: variable is %s', source_data.name)

        # get the gridtype from class that matches the data
        gridtype = self._get_gridtype(datagridtype)

        # if the current grid is not available in the Regrid gridtype
        if gridtype is None:
            self.loggy.info('%s will be excluded from the output', source_data.name)
            return xarray.DataArray(data=None)

        # select the gridtype to be used
        mask_dim = gridtype.mask_dim
        weights = gridtype.weights
        weights_matrix = gridtype.weights_matrix
        masked = gridtype.masked
        horizontal_dims = gridtype.horizontal_dims

        # CDO 2.2.0 fix
        if "numLinks" in weights.dims:
            links_dim = "numLinks"
        else:
            links_dim = "num_links"

        # this is necessary to remove lev-bounds, temporary hack since they should
        # be treated in a smarter way
        if ("bnds" in source_data.name or "bounds" in source_data.name):
            return source_data

        data3d_list = []
        # create pandas object for fast level index retrieval
        mask_index = weights.coords[mask_dim].to_index()
        for idx, lev in enumerate(source_data.coords[mask_dim].values):
            # get the index of the level for weights selection (widx, which might be different from idx)
            # tolerance is necessary to avoid numerical issues, hard coded
            widx = mask_index.get_indexer([lev], method="nearest", tolerance=1e-3)[0]
            if widx == -1:
                raise ValueError(
                    f"{lev} not found in mask_dim {mask_index.name}. "
                    f"Available levels: {list(mask_index.values)}"
                )
            #widx = mask_index.get_loc(lev)
            self.loggy.debug('Processing vertical level %s - level_index %s', lev, widx)
            # use isel since it faster
            xa = source_data.isel(**{mask_dim: idx})
            wa = weights.isel(**{mask_dim: widx})
            self.loggy.debug('Weight number of links is %s', wa.link_length.values)
            nl = wa.link_length.values
            wa = wa.isel(**{links_dim: slice(0, nl)})
            wm = weights_matrix[widx]
            mm = masked[widx]
            data3d_list.append(self.apply_weights(
                xa, wa, weights_matrix=wm,
                masked=mm, horizontal_dims=horizontal_dims)
            )
        data3d = xarray.concat(data3d_list, dim=mask_dim, coords='different', compat='equals')

        # get dimensional info on target grid. TODO: can be moved at the init?
        target_gridtypes = GridInspector(data3d, clean=True, loglevel=self.loglevel).get_gridtype()
        target_horizontal_dims = target_gridtypes[0].horizontal_dims

        if self.transpose:
            data3d = self._transpose_output(data3d, target_horizontal_dims)
        return data3d

    def _transpose_output(self, data3d, target_horizontal_dims):
        """
        Transpose the output data array to place the vertical dimension as first
        """
        dims = list(data3d.dims)
        index = min((i for i, s in enumerate(dims) if s in target_horizontal_dims))
        dimst = dims[1:index] + [dims[0]] + dims[index:]
        return data3d.transpose(*dimst)

    def regrid2d(self, source_data, datagridtype):
        """
        Regrid a 2D source data array to match the target grid.

        This method applies the necessary weights for 2D data regridding.

        Args:
            source_data (xarray.DataArray): The 2D source data to be regridded.
            datagridtype: The grid type information for the source data.

        Returns:
            xarray.DataArray: The regridded 2D data.
        """
        self.loggy.debug('2D DataArray access: variables is %s', source_data.name)

        # get the gridtype from class that matches the data
        gridtype = self._get_gridtype(datagridtype)

        if gridtype is None:
            self.loggy.info('%s will be excluded from the output', source_data.name)
            return xarray.DataArray(data=None)

        return self.apply_weights(
            source_data,
            gridtype.weights,
            weights_matrix=gridtype.weights_matrix,
            masked=gridtype.masked,
            horizontal_dims=gridtype.horizontal_dims)

    def apply_weights(self, source_data, weights, weights_matrix=None,
                      masked=True, horizontal_dims=None):
        """
        Apply CDO weights to the source data, performing the regridding operation.

        This method multiplies the source data with the weights matrix to produce
        the regridded output.

        Args:
            source_data (xarray.DataArray): The source data to be regridded.
            weights (xarray.DataArray): The CDO weights for interpolation.
            weights_matrix (xarray.DataArray, optional): Pre-computed weights matrix (default: None).
            masked (bool): Indicates if the DataArray is masked (default: True).
            horizontal_dims (list): List of dimensions for interpolation (e.g., ['lon', 'lat']).

        Returns:
            xarray.DataArray: The regridded version of the source dataset.

        Raises:
            KeyError: If none of the specified horizontal dimensions are found in the DataArray.
        """

        # Understand immediately if we need to return something or not
        # This is done if we have bounds variables
        if any(substring in source_data.name for substring in ["bnds", "bounds", "vertices"]):

            # we keep time bounds, and we ignore all the rest
            if 'time' in source_data.name:
                self.loggy.info('%s will not be interpolated in the output', source_data.name)
                return source_data

            self.loggy.info('%s will be excluded from the output', source_data.name)
            return xarray.DataArray(data=None)

        # CDO style weights
        # src_address = w.src_address - 1
        # dst_address = w.dst_address - 1
        # remap_matrix = w.remap_matrix[:, 0]
        # w_shape = (w.sizes["src_grid_size"], w.sizes["dst_grid_size"])
        # src_grid_rank = w.src_grid_rank

        # info on grids
        src_cdo_grid = weights.attrs['source_grid']
        dst_cdo_grid = weights.attrs['dest_grid']
        self.loggy.info('Interpolating from CDO %s to CDO %s', src_cdo_grid, dst_cdo_grid)

        # src grid properties
        src_grid_shape = weights.src_grid_dims.values

        # destination grid properties
        dst_grid_shape = weights.dst_grid_dims.values
        dst_frac_area = weights.dst_grid_frac
        dst_grid_mask = weights.dst_grid_imask
        dst_grid_rank = weights.dst_grid_rank

        # target lon/lat
        dst_grid_center_lat = weights.dst_grid_center_lat.data.reshape(dst_grid_shape[::-1])
        dst_grid_center_lon = weights.dst_grid_center_lon.data.reshape(dst_grid_shape[::-1])

        axis_scale = 180.0 / math.pi  # Weight lat/lon in radians

        if not any(x in source_data.dims for x in horizontal_dims):
            self.loggy.error(
                "None of dimensions on which we can interpolate is found in the DataArray. Does your DataArray include any of these?")
            self.loggy.error(horizontal_dims)
            self.loggy.error('smmregrid can identify only %s', source_data.dims)
            raise KeyError('Dimensions mismatch')

        self.loggy.info('Regridding from %s to %s', src_grid_shape, dst_grid_shape)

        # Find dimensions to keep
        kept_dims = [dim for dim in source_data.dims if dim not in horizontal_dims]
        kept_shape = [source_data.sizes[dim] for dim in kept_dims]

        self.loggy.debug('Dimension to be ignored: %s', kept_dims)
        if weights_matrix is None:
            weights_matrix = compute_weights_matrix(weights)

        # Remove the spatial axes, apply the weights, add the spatial axes back
        source_array = source_data.data
        if isinstance(source_array, dask.array.Array):
            source_array = dask.array.reshape(source_array, kept_shape + [-1])
        else:
            source_array = numpy.reshape(source_array, kept_shape + [-1])
        self.loggy.debug('Source array after reshape is: %s', source_array.shape)

        # Handle input mask
        dask.array.ma.set_fill_value(source_array, 1e20)
        source_array = dask.array.ma.fix_invalid(source_array)
        source_array = dask.array.ma.filled(source_array)

        self.loggy.debug('Tensordot!')
        target_dask = dask.array.tensordot(source_array, weights_matrix, axes=1)

        # define and compute the new mask
        if masked:
            # broadcast the mask on all the remaining dimensions and apply it with where
            # Reshape mask to broadcastable shape (avoid intermediate boolean array creation)
            mask_shape = [1 for d in kept_shape] + [-1]
            target_mask = dst_grid_mask.data.reshape(mask_shape).astype(bool)
            self.loggy.debug('Applying mask with shape %s', target_dask.shape)
            target_dask = dask.array.where(target_mask, target_dask, numpy.nan)

        # use the frac area of the destination to further mask the data
        if self.remap_area_min > 0.0:
            target_dask = dask.array.where(
                dask.array.broadcast_to(dst_frac_area, target_dask.shape) < self.remap_area_min,
                numpy.nan, target_dask)

        # after the tensordot, bring the NaN back in
        # Use greater than 1e19 to avoid numerical noise from interpolation.
        # TODO: verify if after mask we still need this
        target_dask = dask.array.where(target_dask > 1e19, numpy.nan, target_dask)

        if len(dst_grid_rank) == 2:
            tgt_shape = [dst_grid_shape[1], dst_grid_shape[0]]
            tgt_dims = ["i", "j"]
        elif len(dst_grid_rank) == 1:
            tgt_shape = [dst_grid_shape[0]]
            tgt_dims = ['cell']
        else:
            raise ValueError('Unknown dimensional target grid')

        # reshape the target DataArray
        target_dask = dask.array.reshape(
            target_dask, kept_shape + tgt_shape
        )

        # Create a new DataArray for the output
        target_da = xarray.DataArray(
            target_dask,
            dims=kept_dims + tgt_dims,
            coords={
                k: v
                for k, v in source_data.coords.items()
                if set(v.dims).issubset(kept_dims)
            },
            name=source_data.name,
        )

        # Add the destination grid coordinate
        target_da.coords["lat"] = xarray.DataArray(dst_grid_center_lat, dims=tgt_dims)
        target_da.coords["lon"] = xarray.DataArray(dst_grid_center_lon, dims=tgt_dims)

        # Clean up coordinates
        target_da.coords["lat"] = remove_degenerate_axes(target_da.lat)
        target_da.coords["lon"] = remove_degenerate_axes(target_da.lon)

        # Convert to degrees if needed, rounding to avoid numerical errors
        target_da.coords["lat"] = numpy.round(target_da.lat * axis_scale, 10)
        target_da.coords["lon"] = numpy.round(target_da.lon * axis_scale, 10)

        # If a regular grid drop the 'i' and 'j' dimensions
        if tgt_dims == ['i', 'j'] and target_da.coords["lat"].ndim == 1 and target_da.coords["lon"].ndim == 1:
            target_da = target_da.swap_dims({"i": "lat", "j": "lon"})

        # Add metadata to the coordinates
        target_da.coords["lat"].attrs.update(
            {"units": "degrees_north", "standard_name": "latitude", "axis": "Y"}
        )
        target_da.coords["lon"].attrs.update(
            {"units": "degrees_east", "standard_name": "longitude", "axis": "X"}
        )

        # Copy attributes from the original
        target_da.attrs = source_data.attrs

        # Clean CDI gridtype (which can lead to issues with CDO interpretation)
        target_da.attrs.pop('CDI_grid_type', None)

        return target_da

    def _check_nan_variation(self, source_grid, gridtype):
        """
        Check for NaN variation in the source grid and update the gridtype if necessary.
        """
        # check for NaN variation in the source grid
        if gridtype.mask_dim:
            self.loggy.info('Gridtype already has a masked dimension %s', gridtype.mask_dim)
            return gridtype

        var = next(iter(gridtype.variables))
        self.loggy.info('Checking for NaN variation in the source grid for variable %s', var)
        nan_dims = detect_nan_variation_dims(
            source_grid[var], time_dim=gridtype.time_dims,
            check_dims=gridtype.other_dims
        )

        if not nan_dims:
            return gridtype

        self.loggy.warning('Found NaN variation in dimensions: %s', nan_dims)
        # update extra dimension for future calls
        self.extra_dims['mask'] = nan_dims
        return GridInspector(source_grid, extra_dims=self.extra_dims,
                             loglevel=self.loglevel, clean=True).get_gridtype()[0]


def regrid(source_data, target_grid=None, weights=None, transpose=True, cdo='cdo'):
    """
    A simple regrid. Inefficient if you are regridding more than one dataset
    to the target grid because it re-generates the weights each time you call
    the function.

    To save the weights use :class:`Regridder`.

    Args:
        source_data (:class:`xarray.DataArray`): Source variable
        target_grid (:class:`coecms.grid.Grid` or :class:`xarray.DataArray`): Target grid / sample variable
        weights (:class:`xarray.Dataset`): Pre-computed interpolation weights
        transpose (bool): If True, transpose the output so that the masked
                          coordinate is just before the other spatial coordinates (default: True)

        cdo (path): path of cdo executable ["cdo"]
    Returns:
        :class:`xarray.DataArray` with a regridded version of the source variable
    """

    regridder = Regridder(source_data, target_grid=target_grid, weights=weights, cdo=cdo, transpose=transpose)
    return regridder.regrid(source_data)



""""
Class to define CDO objects and their grid types.
This module provides a class to represent CDO grid strings and
validate them against known CDO grid patterns.
"""

import re

# Define CDO regex patterns for each grid type (updated at CDO 2.4.4)
# Define regex patterns for each grid type
CDO_GRID_PATTERNS = {
    "global_regular": re.compile(r"^global_\d+(\.\d+)?$"),
    "regional_regular": re.compile(r"^dcw:[A-Z]{2,4}(?:_\d+(\.\d+)?)?$"),
    "zonal_latitudes": re.compile(r"^zonal_\d+(\.\d+)?$"),
    "global_regular_NxM": re.compile(r"^r\d+x\d+$"),
    "one_grid_point": re.compile(r"^lon=(-?\d+(\.\d+)?)/lat=(-?\d+(\.\d+)?)$"),
    "gaussian_grid_F": re.compile(r"^F\d+$"),
    "gaussian_grid_n": re.compile(r"^n\d+$"),
    "icosahedral_gme": re.compile(r"^gme\d+$"),
    "healpix_grid": re.compile(r"^hp\d+(?:_(nested|ring))?$"),
    "healpix_zoom": re.compile(r"^hpz\d+$")
}


class CdoGrid:
    """Class to represent and validate a CDO grid string."""

    def __init__(self, grid_str):
        if not isinstance(grid_str, str):
            raise TypeError("CDOGrid must be initialized with a string.")
        self.grid_kind = self._is_cdo_grid(grid_str)
        if self.grid_kind is None:
            self.grid_str = 'Invalid'
        else:
            self.grid_str = grid_str

    def _is_cdo_grid(self, grid_str):
        """Check if the input string matches any CDO grid type."""

        # Check if the string matches any of the grid patterns
        for grid_type, pattern in CDO_GRID_PATTERNS.items():
            if pattern.match(grid_str):
                return grid_type

        # Return False if no patterns match
        return None

    def __repr__(self):
        return f"CDOGrid(grid_str='{self.grid_str}', grid_kind='{self.grid_kind}')"



"""smmregrid module"""

from .regrid import Regridder, regrid
from .cdogenerate import cdo_generate_weights, CdoGenerate
from .cdogrid import CdoGrid
from .gridtype import GridType
from .gridinspector import GridInspector


__version__ = '0.1.5'

__all__ = ['Regridder', 'regrid', 'cdo_generate_weights',
           'GridType', 'GridInspector', 'CdoGenerate',
           'CdoGrid']



#!/usr/bin/env python
# Copyright 2018 ARC Centre of Excellence for Climate Extremes
# author: Scott Wales <scott.wales@unimelb.edu.au>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

# from cfunits import Units
import numpy


def remove_degenerate_axes(coord):
    """
    Remove any degenerate axes from the coordinate, where all the values along a dimension are identical

    Args:
        coord (xarray.DataArray): Co-ordinate to operate on

    Returns:
        xarray.DataArray with degenerate axes removed
    """

    for d in coord.dims:
        if numpy.allclose(coord.max(dim=d) - coord.min(dim=d), 0):
            coord = coord.mean(dim=d)

    return coord


# def identify_lat_lon(dataarray):
#     """
#     Identify the latitude and longitude dimensions of a dataarray using CF
#     attributes

#     Args:
#         dataarray: Source dataarray

#     Returns:
#         (lat, lon): Tuple of `xarray.Dataarray` for the latitude and longitude
#             dimensions

#     Todo:
#         * Assumes latitude and longitude are unique
#     """

#     lat = None
#     lon = None

#     for c in dataarray.coords.values():
#         if (
#             c.attrs.get("standard_name", "") == "latitude"
#             or Units(c.attrs.get("units", "")).islatitude
#             or c.attrs.get("axis", "") == "Y"
#         ):
#             lat = c

#         if (
#             c.attrs.get("standard_name", "") == "longitude"
#             or Units(c.attrs.get("units", "")).islongitude
#             or c.attrs.get("axis", "") == "X"
#         ):
#             lon = c

#     if lat is None or lon is None:
#         raise Exception("Couldn't identify horizontal coordinates")

#     return (lat, lon)


# def identify_time(dataarray):
#     """
#     Identify the time dimension of a dataarray using CF attributes

#     Args:
#         dataarray: Source dataarray

#     Returns:
#         :obj:`xarray.Dataarray` for the time dimension

#     Todo:
#         * Assumes time dimension is unique
#     """

#     for c in dataarray.coords.values():
#         if (
#             c.attrs.get("standard_name", "") == "time"
#             or Units(c.attrs.get("units", "")).isreftime
#             or Units(c.encoding.get("units", "")).isreftime
#             or c.attrs.get("axis", "") == "T"
#         ):
#             return c

#     raise Exception("No time axis found")



"""Module for cdo-smmregrid comparison"""

import numpy as np
import xarray as xr
from cdo import Cdo
from smmregrid import Regridder, CdoGenerate
cdo = Cdo()


def find_var(xfield):
    """Find them most likely set of vars that needs to be interpolated"""

    # var as the one which have time and not have bnds (could work)
    myvar = [var for var in xfield.data_vars
             if 'time' in xfield[var].dims and 'bnds' not in xfield[var].dims]

    # if find none otherwise, pick what you have
    if not myvar:
        myvar = list(xfield.data_vars)

    return myvar


def check_cdo_regrid(finput, ftarget, remap_method='con', access='Dataset',
                     init_method='grids', mask_dim=None, vertical_dim=None, extrapolate=True,
                     remap_area_min=0.0, loglevel='INFO'):
    """Given a file to be interpolated finput over the ftarget grid,
    check if the output of the last variable is the same as produced
    by CDO remap command. This function is used for tests."""

    # define files and open input file
    if isinstance(finput, str):
        xfield = xr.open_mfdataset(finput)
    else:
        xfield = finput

    # convert extrapolation to cdo!
    cdoextrapolate = 'on' if extrapolate else 'off'

    # interpolation with pure CDO
    cdo_interpolator = getattr(cdo, 'remap' + remap_method)
    cdofield = cdo_interpolator(ftarget, input=finput, returnXDataset=True,
                                env={'REMAP_EXTRAPOLATE': cdoextrapolate})
    # 'REMAP_AREA_MIN': remap_area_min}) # this is not working with cdo 2.4.4

    # var as the one which have time and not have bnds (could work)
    smmvar = find_var(xfield)
    cdovar = find_var(cdofield)

    # method with automatic creation of weights
    if init_method == 'grids':
        interpolator = Regridder(source_grid=finput, target_grid=ftarget, 
                                 remap_area_min=remap_area_min,
                                 method=remap_method, mask_dim=mask_dim, 
                                 vertical_dim=vertical_dim, loglevel=loglevel)
    elif init_method == 'weights':
        wfield = CdoGenerate(finput, ftarget, loglevel=loglevel).weights(
            method=remap_method, mask_dim=mask_dim, vertical_dim=vertical_dim)
        interpolator = Regridder(weights=wfield, loglevel=loglevel, remap_area_min=remap_area_min)
    else:
        raise KeyError('Unsupported init method')
    rfield = interpolator.regrid(xfield)

    if access == 'Dataset':
        rfield = rfield[smmvar].to_array()
        cdofield = cdofield[cdovar].to_array()

    # check if arrays are equal with numerical tolerance
    checker = np.allclose(cdofield, rfield, equal_nan=True)
    return checker


def check_cdo_regrid_levels(finput, ftarget, mask_dim, levels, remap_method='con',
                            remap_area_min=0.5, access='Dataset',
                            extrapolate=True, loglevel='INFO'):
    """Given a file to be interpolated finput over the ftarget grid,
    check if the output of the last variable is the same as produced
    by CDO remap command. This function is used for tests.
    This is a variant to check the level memory feature (idx_3d)."""

    # define files and open input file
    xfield = xr.open_mfdataset(finput)

    # convert extrapolate to CDO
    cdoextrapolate = 'on' if extrapolate else 'off'

    # interpolation with pure CDO
    cdo_interpolator = getattr(cdo, 'remap' + remap_method)
    cdofield = cdo_interpolator(ftarget, input=finput, returnXDataset=True,
                                env={'REMAP_EXTRAPOLATE': cdoextrapolate,
                                     'REMAP_AREA_MIN': str(remap_area_min)})

    # Keep only some levels
    cdofield = cdofield.isel(**{mask_dim: levels})

    # var as the one which have time and not have bnds (could work)
    smmvar = find_var(xfield)
    cdovar = find_var(cdofield)

    # compute weights
    wfield = CdoGenerate(finput, ftarget, loglevel=loglevel).weights(
        method=remap_method, mask_dim=mask_dim)

    # Pass full 3D weights
    interpolator = Regridder(weights=wfield, loglevel=loglevel, remap_area_min=remap_area_min)

    # subselect some levels
    xfield = xfield.isel(**{mask_dim: levels})

    # Regrid level selection
    rfield = interpolator.regrid(xfield)

    if access == 'Dataset':
        rfield = rfield[smmvar].to_array()
        cdofield = cdofield[cdovar].to_array()

    # check if arrays are equal with numerical tolerance
    checker = np.allclose(cdofield, rfield, equal_nan=True)
    return checker



"""Module to implement logging configurations"""
import logging


def setup_logger(level=None, name=None):
    """Define a logger to be used extensively within smmregrid"""

    if name is None:
        name = "smmregrid"

    loglev = convert_logger(level)

    logger = logging.getLogger(name)  # Create a logger specific to your module

    if logger.handlers:
        # logger.warning('Logging is already setup with name %s', name)
        if level != logging.getLevelName(logger.getEffectiveLevel()):
            logger.setLevel(loglev)
            logger.info('Updating the log_level to %s', loglev)
        return logger

    # avoid duplication/propagation of loggers
    logger.propagate = False

    logger.setLevel(loglev)  # Set the desired log level

    # Create a formatter to specify the log format
    formatter = logging.Formatter('%(asctime)s | %(name)s | %(levelname)8s -> %(message)s',
                                  datefmt='%Y-%m-%d %H:%M:%S')

    # Create a handler for the logger
    handler = logging.StreamHandler()
    # handler.setLevel(loglev)  # Set the desired log level for the handler
    handler.setFormatter(formatter)  # Assign the formatter to the handler
    logger.addHandler(handler)  # Add the handler to the logger

    return logger


def convert_logger(loglev=None):
    """Convert a string or integer to a valid logging level"""

    # Default log level for the AQUA framework
    loglev_default = "WARNING"

    # If loglev is a string, convert it to uppercase
    if isinstance(loglev, str):
        loglev = loglev.upper()

    # If loglev is an integer, convert it to a string
    elif isinstance(loglev, int):
        loglev = logging.getLevelName(loglev)

    # If loglev is None, set it to the default log level
    elif loglev is None:
        loglev = loglev_default

    # If loglev is of an unsupported type, raise a ValueError
    else:
        raise ValueError('Invalid log level type. Must be a string or an integer.')

    # Check if the log level exists and retrieve its integer value
    loglev_int = getattr(logging, loglev, None)

    # If loglev_int is None, the log level doesn't exist
    if loglev_int is None:
        logging.warning("Invalid logging level '%s' specified. Setting it back to default '%s'.",
                        loglev, loglev_default)
        loglev = loglev_default

    return loglev



"""CDO-based generation of weights"""

import os
import sys
import copy
import tempfile
import subprocess
import warnings
from multiprocessing import Process, Manager
import numpy
import xarray
from .weights import compute_weights_matrix3d, compute_weights_matrix, mask_weights, check_mask
from .log import setup_logger
from .cdogrid import CdoGrid
from .util import deprecated_argument, tolist


class CdoGenerate():
    """CDO-based class to generate weights for smmregrid."""

    def __init__(self, source_grid, target_grid=None, cdo_extra=None,
                 cdo_options=None, cdo_download_path=None, cdo_icon_grids=None,
                 cdo="cdo", loglevel='warning'):
        """
        Initialize GenerateWeights class for regridding using Climate Data Operators (CDO),
        accommodating both 2D and 3D grid cases.

        Args:
            source_grid (str or xarray.Dataset): The source grid from which to generate weights.
                                                 This can be a file path or an xarray dataset.
            target_grid (str or xarray.Dataset): The target grid to which the source grid
                                                 will be regridded. This can also be a file
                                                 path or an xarray dataset.
            loglevel (str, optional): The logging level for messages. Default is 'warning'.
                                      Options include 'debug', 'info', 'warning',
                                      'error', and 'critical'.
            cdo_extra (list, optional): Additional CDO command-line options.
                                               Defaults to None.
            cdo_options (list, optional): Options for CDO commands. Defaults to None.
            cdo (str, optional): The command to invoke CDO. Default is "cdo".
            cdo_icon_grids (str, optional): Path to the ICON grid
                                            if applicable. Defaults to None.
            cdo_download_path (str, optional): Path to the grid download path
                                                if applicable. Defaults to None.
        """

        self.loggy = setup_logger(level=loglevel, name='smmregrid.CdoGenerate')

        # cdo options and extra, ensure they are lists
        self.cdo_extra = tolist(cdo_extra)
        self.cdo_options = tolist(cdo_options)

        # assign the two grids
        self.source_grid = source_grid
        self.target_grid = target_grid

        # define grid filenames
        self.source_grid_filename = None
        self.target_grid_filename = None

        # assign cdo bin file and get envrinment file
        self.cdo = cdo
        self.env = os.environ.copy()
        if cdo_download_path:
            self.env["CDO_DOWNLOAD_PATH"] = cdo_download_path
        if cdo_icon_grids:
            self.env["CDO_ICON_GRIDS"] = cdo_icon_grids

    @staticmethod
    def _safe_check(method, remap_norm):
        """Safety checks for weights generation """

        supported_methods = ["bic", "bil", "con", "con2", "dis", "laf", "nn", "ycon"]
        if method not in supported_methods:
            raise ValueError('The remap method provided is not supported!')
        if remap_norm not in ["fracarea", "destarea"]:
            raise ValueError('The remap normalization provided is not supported!')

    def _prepare_grid(self, grid, target=False):
        """Helper function to prepare grid (file or dataset)."""

        # if grid is a Dataset or DataArray, save it to a temporary file
        if isinstance(grid, (xarray.Dataset, xarray.DataArray)):
            grid_file = tempfile.NamedTemporaryFile(delete=False)
            grid.to_netcdf(grid_file.name)
            self.loggy.debug("Xarray source, temporary grid file created for areas/weights generation: %s", grid_file.name)
            return grid_file.name

        # prepare grid: if source_grid is a CDO grid, use it as is, otherwise prepare the file
        if CdoGrid(grid).grid_kind and not target:
            self.loggy.info('CDO grid as %s to be used for areas/weights generation', grid)
            return f"-const,1,{grid}"

        # if grid is a string, assume it's a file path
        if isinstance(grid, str):
            self.loggy.debug("Grid file path to be used for areas/weights generation: %s", grid)
            return grid

        raise TypeError('Grid must be a CDO grid string, a file path, or an xarray Dataset/DataArray.')

    def weights(self, method="con", extrapolate=True,
                remap_norm="fracarea",
                mask_dim=None, vertical_dim=None, nproc=1):
        """
        Generate weights for regridding using Climate Data Operators (CDO),
        accommodating both 2D and 3D grid cases.

        Args:
            method (str, optional): The remapping method to use.
                                    Default is "con" for conservative remapping.
                                    Other options may include 'bil', 'nearest', etc.
            extrapolate (bool, optional): Whether to allow extrapolation beyond the grid boundaries.
                                          Defaults to True.
            remap_norm (str, optional): The normalization method to apply when remapping.
                                        Default is "fracarea" which normalizes by fractional area.
            nproc (int, optional): Number of processes to use for parallel processing. Default is 1.
            mask_dim (str, optional): Name of the vertical dimension in the source grid, if applicable.
                                        Defaults to None. Use if the grid is 3D.
            vertical_dim (str, optional): Deprecated. Use `mask_dim` instead. Defaults to None.

        Returns:
            xarray.Dataset: A dataset containing the generated weights
                            and a mask indicating which grid cells
                            were successfully masked.
                            The mask is stored in a variable named "dst_grid_masked".

        Raises:
            KeyError: If the specified vertical dimension cannot be found in the source grid.

        Notes:
            This function handles both 2D and 3D grid cases:

            For 2D grids (when `mask_dim` is None), it calls the `_weights_2d` method
            to generate weights. The weights are then masked based on a precomputed weights matrix.

            For 3D grids (when `mask_dim` is specified), it uses multiprocessing to generate weights
            for each vertical level. It requires the vertical dimension to be present in the source grid,
            and it will generate a mask indicating valid and invalid weights for each vertical level.

            The function logs the progress of weight generation, including the length of vertical dimensions
            and each level being processed.

            Deprecation Warning: The `vertical_dim` parameter is deprecated and will be removed in future versions.
            Users should migrate to using `mask_dim`, respectively.

        """
        if self.target_grid is None:
            raise TypeError('Target grid is not specified, cannot provide any regridding')
        if self.source_grid is None:
            raise TypeError('Source grid is not specified, cannot provide any regridding')

        # verify that method and normalization are suitable
        self._safe_check(method, remap_norm)

        # vertical dimension
        mask_dim = deprecated_argument(vertical_dim, mask_dim, 'vertical_dim', 'mask_dim')

        # prepare grid
        self.source_grid_filename = self._prepare_grid(self.source_grid)
        self.target_grid_filename = self._prepare_grid(self.target_grid, target=True)

        # Generate weights for 2D or 3D grid based on mask_dim presence
        if not mask_dim:
            return self._weights_2d(method, extrapolate, remap_norm)
        return self._weights_3d(method, extrapolate, remap_norm,
                                nproc, mask_dim)

    def _weights_2d(self, method, extrapolate, remap_norm):
        """Generate 2D weights using CDO."""

        weights = self._cdo_generate_weights(method, extrapolate,
                                             remap_norm)
        weights_matrix = compute_weights_matrix(weights)
        weights = mask_weights(weights, weights_matrix)
        masked = int(check_mask(weights))
        masked_xa = xarray.DataArray(masked, name="dst_grid_masked")
        return xarray.merge([weights, masked_xa])

    def _weights_3d(self, method, extrapolate, remap_norm,
                    nproc, mask_dim):
        """Generate 3D weights using multiprocessing."""

        if isinstance(self.source_grid, str):
            sgrid = xarray.open_dataset(self.source_grid)
        else:
            sgrid = self.source_grid

        if mask_dim not in sgrid.dims:
            raise KeyError(f'Cannot find vertical dim {mask_dim} in {list(sgrid.dims)}')

        # nvert = sgrid[mask_dim].values.size
        nvert = sgrid.sizes[mask_dim]
        self.loggy.info('Vertical dimension has length: %s', nvert)

        mgr = Manager()
        wlist = mgr.list(range(nvert))

        num_blocks, remainder = divmod(nvert, nproc)
        num_blocks += 0 if remainder == 0 else 1

        blocks = numpy.array_split(numpy.arange(nvert), num_blocks)
        for block in blocks:
            processes = []
            for lev in block:
                self.loggy.info("Generating level: %s at depth %s", str(lev), sgrid[mask_dim].values[lev])
                cdo_extra_vertical = [f"-sellevidx,{lev + 1}"]
                ppp = Process(target=self._weights_worker,
                              args=(wlist, lev),
                              kwargs={
                                  "method": method,
                                  "extrapolate": extrapolate,
                                  "remap_norm": remap_norm,
                                  "cdo_extra_vertical": cdo_extra_vertical
                              })
                ppp.start()
                processes.append(ppp)
            for proc in processes:
                proc.join()

        weights = self.weightslist_to_3d(wlist, method, sgrid[mask_dim])
        weights_matrix = compute_weights_matrix3d(weights, mask_dim)
        weights = mask_weights(weights, weights_matrix, mask_dim)
        masked = check_mask(weights, mask_dim)
        masked_xa = xarray.DataArray(masked,
                                     coords={mask_dim: weights[mask_dim]},
                                     name="dst_grid_masked")

        return xarray.merge([weights, masked_xa])

    def _weights_worker(self, wlist, nnn, *args, **kwargs):
        """Run a worker process."""
        wlist[nnn] = self._cdo_generate_weights(*args, **kwargs).compute()

    def _cdo_generate_weights(self, method="con", extrapolate=True,
                              remap_norm="fracarea",
                              cdo_extra_vertical=None):
        """
        Generate weights for regridding using CDO

        Available weight generation methods are:

        * bic: SCRIP Bicubic
        * bil: SCRIP Bilinear
        * con: SCRIP First-order conservative
        * con2: SCRIP Second-order conservative
        * dis: SCRIP Distance-weighted average
        * laf: YAC Largest area fraction
        * ycon: YAC First-order conservative
        * nn: Nearest neighbour

        Args:
            method (str): Regridding method - default "con"
            extrapolate (bool): Extrapolate output field - default True
            remap_norm (str): Normalisation method for conservative methods
            cdo_extra_vertical (list): Command to select vertical levels for 3d case

        Returns:
            :obj:`xarray.Dataset` with regridding weights
        """

        sgrid = self.source_grid_filename
        tgrid = self.target_grid_filename

        cdo_extra_vertical = tolist(cdo_extra_vertical)

        # Log method and remapping information
        self.loggy.info("CDO remapping method: %s", method)
        self.loggy.info("Extrapolation enabled: %s", extrapolate)
        self.loggy.debug("Normalization method: %s", remap_norm)

        weight_file = tempfile.NamedTemporaryFile()
        self.loggy.debug("Weight file name is: %s", weight_file.name)

        self.loggy.debug("Source grid file name is: %s", sgrid)
        self.loggy.debug("Target grid file name is: %s", tgrid)

        env = copy.deepcopy(self.env)
        env["REMAP_EXTRAPOLATE"] = "on" if extrapolate else "off"
        env["CDO_REMAP_NORM"] = remap_norm

        try:
            self.loggy.info("Additional CDO commands: %s", self.cdo_extra)
            self.loggy.info("Additional CDO options: %s", self.cdo_options)

            command = [
                self.cdo,
                *self.cdo_options,
                f"gen{method},{tgrid}",
                *self.cdo_extra + cdo_extra_vertical,
                sgrid,
                weight_file.name
            ]
            self.loggy.debug("Final CDO command: %s", command)
            subprocess.check_output(command, stderr=subprocess.STDOUT, env=env)

            weights = xarray.open_dataset(weight_file.name, engine="netcdf4")
            return weights

        except subprocess.CalledProcessError as err:
            print(err.output.decode(), file=sys.stderr)
            raise
        finally:
            weight_file.close()

    # def _remove_tmpfile(self, tmpfile):
    #     """Helper function to clean the tempfile grid (file or dataset)."""
    #     if not isinstance(tmpfile, str):
    #         os.remove(tmpfile.name)

    def weightslist_to_3d(self, weights_list, method='ycon', mask_coord=xarray.DataArray):
        """Combine a list of 2D CDO weights into a 3D one.
        
        Args:
            ds_list (list): List of xarray.Dataset with 2D weights
            method (str): Remap method, to determine if dst_grid_area and dst_grid_frac
                          need to be included
            mask_coord (xarray.DataArray): Coordinate to use for the 3D weights
        """

        links_dim = "numLinks" if "numLinks" in weights_list[0].dims else "num_links"
        number_links = [ds.src_address.size for ds in weights_list]
        links_array = xarray.DataArray(number_links, coords={mask_coord.name: mask_coord.values}, name="link_length")

        varlist = ["src_address", "dst_address", "remap_matrix", "src_grid_imask", "dst_grid_imask"]
        # Add dst_grid_area and dst_grid_frac only for conservative methods
        if method in ['ycon', 'con2', 'con']:
            varlist += ["dst_grid_area", "dst_grid_frac"]
        untouched_array = weights_list[0].drop_vars(varlist)

        modified_array  = []
        nl_max = max(number_links)
        for x, v in zip(weights_list, mask_coord.values):
            nl1 = x.src_address.size
            xplist = [x[vname].pad(**{links_dim: (0, nl_max - nl1), "mode": 'constant', "constant_values": 0})
                      for vname in varlist]
            xmerged = xarray.merge(xplist)
            modified_array.append(xmerged.assign_coords({mask_coord.name: v}))
        modified_array =  xarray.concat(modified_array, mask_coord.name, coords='different', compat='equals')

        merged_array = xarray.merge([links_array, untouched_array, modified_array],
                            combine_attrs='no_conflicts')
        merged_array[mask_coord.name].attrs = mask_coord.attrs
        return merged_array
    
    def areas(self, target=False):
        """Generate source areas or target areas"""

        if not target:
            self.loggy.info('Generating areas for source grid!')
            # if not self.source_grid:
            #    raise TypeError('Source grid is not specified, cannot provide any area')
            return self._areas(self.source_grid, cdo_extra=self.cdo_extra,
                               cdo_options=self.cdo_options)

        if self.target_grid:

            self.loggy.info('Generating areas for target grid!')
            return self._areas(self.target_grid)

        raise TypeError('Target grid is not specified, cannot provide any area')

    def _areas(self, filename, cdo_extra=None, cdo_options=None):
        """Generate areas in a similar way of what done for weights"""

        # safety listing
        cdo_extra = tolist(cdo_extra)
        cdo_options = tolist(cdo_options)

        # Make some temporary files that we'll feed to CDO
        areas_file = tempfile.NamedTemporaryFile()

        # prepare grid
        sgrid = self._prepare_grid(filename)

        try:
            self.loggy.info("Additional CDO commands: %s", cdo_extra)
            self.loggy.info("Additional CDO options: %s", cdo_options)

            command = [
                self.cdo,
                *cdo_options + ["-f", "nc4"],
                "gridarea",
                *cdo_extra,
                sgrid,
                areas_file.name
            ]
            self.loggy.debug("Final CDO command: %s", command)
            subprocess.check_output(command, stderr=subprocess.STDOUT, env=self.env)

            areas = xarray.open_dataset(areas_file.name, engine="netcdf4")
            areas.cell_area.attrs['units'] = 'm2'
            areas.cell_area.attrs['standard_name'] = 'area'
            areas.cell_area.attrs['long_name'] = 'area of grid cell'
            return areas

        except subprocess.CalledProcessError as err:
            print(err.output.decode(), file=sys.stderr)
            raise
        finally:
            areas_file.close()


def cdo_generate_weights(source_grid, target_grid, method="con", extrapolate=True,
                         remap_norm="fracarea", gridpath=None,
                         icongridpath=None,  cdo_extra=None, cdo_options=None,
                         mask_dim=None, vertical_dim=None,
                         cdo="cdo", nproc=1, loglevel='warning'):
    """
    Wrapper function for the new cdo class to provide the old access
    """
    warnings.warn("cdo_generate_weights() is now deprecated, please use CdoGenerate().weights()",
                  DeprecationWarning)
    generator = CdoGenerate(source_grid=source_grid, target_grid=target_grid, loglevel=loglevel,
                            cdo_extra=cdo_extra, cdo_options=cdo_options, cdo=cdo,
                            cdo_icon_grids=icongridpath, cdo_download_path=gridpath)
    return generator.weights(method=method, extrapolate=extrapolate,
                             remap_norm=remap_norm,
                             mask_dim=mask_dim, vertical_dim=vertical_dim, nproc=nproc)



"""Fundamental GridType class to gather all information about grids"""

import copy

# default spatial dimensions and vertical coordinates
DEFAULT_DIMS = {
    'horizontal': ['i', 'j', 'x', 'y', 'lon', 'lat', 'longitude', 'latitude',
                   'cell', 'cells', 'ncells', 'values', 'value', 'nod2', 'pix', 'elem',
                   'nav_lon', 'nav_lat', 'rgrid'],
    'mask': ['lev', 'nz1', 'nz', 'depth', 'depth_full', 'depth_half'],
    'time': ['time', 'time_counter'],
}


class GridType:
    """Fundamental GridType object"""

    def __init__(self, dims, extra_dims=None, override=False, weights=None):
        """
        Initializes a GridType object carrying grid-specific information required by smmregrid.

        Args:
            dims (list): A list of default dimensions for the grid (e.g., ['time', 'lat', 'lon']).
            extra_dims (dict, optional): A dictionary including keys 'mask', 'time', and 'horizontal'
                                          that can be used to extend the default dimensions. Defaults to None.
            override (bool, optional): If True, it will override the default dimensions with the provided ones.
                                         Defaults to False.
            weights (any, optional): CDO weights used in regridding. It will initiate the object in a different
                                         way assuming single-gridtype objects. Defaults to None.

        Attributes:
            horizontal_dims (list): The identified horizontal dimensions from the input.
            mask_dim (str or None): The identified masked dimension, if applicable.
            dims (list): The dimensions defined for the grid. A combination of horizontal and masked dimensions.
            time_dims (list): The identified time dimensions from the input.
            other_dims (list): The dimensions which are there but are not identified automatically.
            variables (dict): A dictionary holding identified variables and their coordinates.
            bounds (list): A list of bounds variables identified in the dataset.
            masked (any): Placeholder for masked data (to be defined later).
            weights (any): The weights associated with the grid, if provided.
            weights_matrix (any): Placeholder for a weights matrix (to be defined later).
        """

        # safety checks
        # if not isinstance(dims, list):
        #    raise TypeError("dims must be a list of dimension names.")
        if extra_dims is not None and not isinstance(extra_dims, dict):
            raise TypeError("extra_dims must be a dictionary or None.")

        default_dims = self._handle_default_dimensions(extra_dims, override=override)
        self.horizontal_dims = self._identify_dims('horizontal', dims, default_dims)
        self.mask_dim = self._identify_dims('mask', dims, default_dims)
        self.dims = (self.horizontal_dims or []) + ([self.mask_dim] if self.mask_dim else [])
        self.time_dims = self._identify_dims('time', dims, default_dims)
        self.other_dims = self._identify_other_dims(dims)

        # used by GridInspector
        self.variables = {}  # dictionary of variables and their coordinates
        self.bounds = []  # list of bounds variables
        self.kind = None  # which kind of grid, regular, guassian, curvilinear, etc.

        # used by Regrid class
        self.masked = None
        self.weights = weights
        self.weights_matrix = None

    def _handle_default_dimensions(self, extra_dims, override=False):
        """
        Extend the default dimensions based on the provided extra dimensions.

        Args:
            extra_dims (dict): A dictionary that can include 'mask', 'time', and
                               'horizontal' keys for extending dimensions.

        Returns:
            dict: An updated dictionary of default dimensions that includes any extensions specified
                  in extra_dims.

        Notes:
            If extra_dims is None, the default dimensions remain unchanged.
        """

        if extra_dims is None:
            return DEFAULT_DIMS

        if override:
            return extra_dims

        # Create a deep copy to avoid modifying the global DEFAULT_DIMS
        update_dims = copy.deepcopy(DEFAULT_DIMS)
        # update_dims = DEFAULT_DIMS

        for dim in extra_dims.keys():
            if extra_dims[dim]:
                update_dims[dim] = list(set(update_dims[dim] + extra_dims[dim]))
        return update_dims

    def __repr__(self):
        """
        Return a string representation of the GridType instance.
        Only includes attributes with values.
        """

        attributes = ', '.join(f"{key}={value}" for key, value in self.__dict__.items() if value is not None)
        return f"{self.__class__.__name__}({attributes})"

    def __eq__(self, other):
        """
        Check for equality between two GridType instances.

        Args:
            other (GridType): Another GridType instance to compare against.

        Returns:
            bool: True if both self.dims are equal for both instances, False otherwise.
        """
        if isinstance(other, GridType):
            return set(self.dims) == set(other.dims)
        return False

    def __hash__(self):
        """
        Generate a hash value for the GridType instance based on its dimensions.

        Returns:
            int: A hash value representing the dimensions of the GridType instance.
        """
        return hash(tuple(self.dims))

    def _identify_dims(self, axis, dims, default_dims):
        """
        Identify dimensions along a specified axis.

        Args:
            axis (str): The axis to check ('horizontal', 'mask', or 'time').
            default_dims (dict): The dictionary of default dimensions to check against.

        Returns:
            list or str: A list of identified dimensions or a single identified masked dimension.
                          Returns None if no dimensions are identified.

        Raises:
            ValueError: If more than one masked dimension is identified.
        """

        # Check if the axis is valid
        if axis not in ['horizontal', 'mask', 'time']:
            raise ValueError(f"Invalid axis '{axis}'. Must be one of 'horizontal', 'mask', or 'time'.")

        # Check if the axis is in the default dimensions
        if axis not in default_dims:
            return None

        # Identify dimensions based on the provided axis
        identified_dims = list(set(dims).intersection(default_dims[axis]))
        if axis == 'mask':
            if len(identified_dims) > 1:
                raise ValueError(f'Only one masked dimension can be processed at the time: check {identified_dims}')
            if len(identified_dims) == 1:
                identified_dims = identified_dims[0]  # unlist the single masked dimension
        return identified_dims if identified_dims else None

    def _identify_other_dims(self, dims):
        """
        Calculate and return the dimensions that are not part of horizontal_dims, mask_dim, or time_dims.

        Returns:
            set: A set of unused dimensions that are not part of horizontal, mask, or time dimensions.
        """
        # Safely convert identified dimensions to sets, handling None values
        horizontal_dims = set(self.horizontal_dims) if self.horizontal_dims else set()
        mask_dim = {self.mask_dim} if self.mask_dim else set()  # Use {} for single value
        time_dims = set(self.time_dims) if self.time_dims else set()

        # Return the unused dimensions by subtracting used dimensions from all dimensions
        return list(set(dims) - (horizontal_dims | mask_dim | time_dims))



"""Utility module with various support function"""

import os
import warnings
import xarray
from .cdogrid import CdoGrid


def check_gridfile(filename):
    """Check if a grid is a file, a cdo string or xarray object"""

    if filename is None:
        return None
    if isinstance(filename, (xarray.Dataset, xarray.DataArray)):
        return "xarray"
    if isinstance(filename, str):
        if CdoGrid(filename).grid_kind:
            return "grid"
        if os.path.exists(filename):
            return "file"
        raise FileNotFoundError(f"Cannot find {filename} on disk")

    raise TypeError(f'Unsuported format for {filename}')


def deprecated_argument(old, new, oldname='var1', newname='var2'):
    """Utility to provide warning in case of deprecated argument"""

    # Check for deprecated 'old' argument
    if old is not None:
        warnings.warn(
            f"{oldname} is deprecated and will be removed in future versions. "
            f"Please use {newname} instead.",
            DeprecationWarning
        )
        # If new is not provided, use the value from old
        if new is None:
            new = old
    return new


def find_coord(ds, possible_names):
    """Find the first matching coordinate in the dataset."""
    return next((name for name in possible_names if name in ds.coords), None)


def tolist(value):
    """Convert value to list format. Returns [] for None, [value] for single values, unchanged for lists."""
    if value is None:
        return []
    if isinstance(value, list):
        return value
    return [value]


def detect_nan_variation_dims(field, time_dim, check_dims):
    """
    Check if NaN values vary along specific dimension(s) in an xarray object
    after removing the time dimension.

    Args:
        field: Input xarray DataArray with dask backend.
        time_dim: Name of the time dimension to remove.
        check_dims:  list[str]. Dimension name(s) to check for NaN variation after time removal.
    
    Returns:
        list[str]: List of dimension names from check_dims that have NaN variations.
    """

    if time_dim[0] in field.dims:
        # If time_dim is present, reduce it to a single value (e.g., first time step)
        field = field.isel({time_dim[0]: 0}, drop=True)
    
    nan_mask = field.isnull()
    dims_with_variation = []
    for dim in check_dims:
        # Variation along this dim
        variation_mask = nan_mask.astype("int8").diff(dim=dim).astype(bool).any(dim=dim)                 
        
        count = variation_mask.sum().compute()
        # print(f"Dimension '{dim}' has {count} variations with NaN values.")
        if count > 0:
            dims_with_variation.append(dim)

    return dims_with_variation



"""GridInspector class module"""

import os
import xarray as xr
import numpy as np
from smmregrid.log import setup_logger
from .gridtype import GridType
from .util import find_coord

# Define coordinate names for latitude and longitude
LAT_COORDS = ["lat", "latitude", "nav_lat"]
LON_COORDS = ["lon", "longitude", "nav_lon"]


class GridInspector():
    """Class to investigate data and detect its GridType() object"""

    def __init__(self, data, cdo_weights=False, extra_dims=None,
                 clean=True, loglevel='warning'):
        """
        GridInspector class to detect information on the data, based on GridType class

        Parameters:
            data (xr.Datase or xr.DataArray): The input dataset.
            clean (bool): apply the cleaning of grids which are assumed to be not relevant
                          for regridding purposes (e.g. bounds)
            cdo_weights (bool): if the data provided are cdo weights instead of data to be regridded
            extra_dims(dict): Extra dimensions to be added and passed to GridType
            loglevel: The loglevel that you want you use
        """
        self.loggy = setup_logger(name='smmregrid.GridInspect', level=loglevel)
        self.loglevel = loglevel
        self.extra_dims = extra_dims
        self.data = self._open_data(data)
        self.cdo_weights = cdo_weights
        self.clean = clean
        self.grids = []  # List to hold all grids info
        self.loggy.debug('Extra_dims are %s', extra_dims)
        self.loggy.debug('Clean flag is %s and cdo_weights init is %s', clean, cdo_weights)

    def _open_data(self, data):
        """
        Open the data from a file path if it is a string and the file exists
        """
        if isinstance(data, (xr.Dataset, xr.DataArray)):
            self.loggy.info('Data is already an xarray Dataset or DataArray')
            return data
        if isinstance(data, str):
            if os.path.exists(data):
                self.loggy.info('Data is a file path, opening xr.dataset')
                return xr.open_dataset(data)
            raise FileNotFoundError(f'File {data} not found')
        raise TypeError('Data supplied is neither xarray Dataset or DataArray')

    def _inspect_grids(self):
        """
        Inspects the dataset and identifies different grids.
        """

        if isinstance(self.data, xr.Dataset):
            self.data.map(self._inspect_dataarray_grid, keep_attrs=False)
        if isinstance(self.data, xr.DataArray):
            self._inspect_dataarray_grid(self.data)

        # get variables associated to the grid
        for gridtype in self.grids:
            self.identify_variables(gridtype)

        # get grid format
        self._identify_grid_format(self.data)

    def _inspect_dataarray_grid(self, data_array):
        """
        Helper method to inspect a single DataArray and identify its grid type.
        If the data_array is a bounds variable, it is not added to the grids list.
        """
        grid_key = tuple(data_array.dims)
        gridtype = GridType(dims=grid_key, extra_dims=self.extra_dims)
        if gridtype not in self.grids and not self._is_bounds(data_array.name):
            self.grids.append(gridtype)

    def _inspect_weights(self):
        """
        Return basic information about CDO weights
        """

        gridtype = GridType(dims=[], weights=self.data)

        # get masked info from the weights coords if available
        if self.data.coords:
            mask_dim = list(self.data.coords)[0]
            self.loggy.debug('Masked dimension read from weights: %s', mask_dim)
            gridtype.mask_dim = mask_dim

        self.grids.append(gridtype)

    def get_gridtype(self):
        """
        Returns detailed information about all the grids in the dataset.
        """

        if self.cdo_weights:
            self.loggy.info('CDO weights are used to define the grid')
            self._inspect_weights()
        else:
            self._inspect_grids()

        if self.clean:
            self._clean_grids()

        # Log details about identified grids
        for gridtype in self.grids:
            self._log_grid_details(gridtype)

        return self.grids

    def _log_grid_details(self, gridtype):
        """Log detailed information about a grid."""
        self.loggy.debug('Grid details: %s', gridtype.dims)
        if gridtype.horizontal_dims:
            self.loggy.debug('  Horizontal dims: %s', gridtype.horizontal_dims)
        if gridtype.mask_dim:
            self.loggy.debug('  Masked dim: %s', gridtype.mask_dim)
        if gridtype.time_dims:
            self.loggy.debug('  Time dims: %s', gridtype.time_dims)
        if gridtype.other_dims:
            self.loggy.debug('  Other dims: %s', gridtype.other_dims)
        self.loggy.debug('  Variables: %s', list(gridtype.variables.keys()))
        self.loggy.debug('  Bounds: %s', gridtype.bounds)

    def _clean_grids(self):
        """
        Remove degenerate grids which are used by not relevant variables
        """
        removed = []

        # Iterate through grids
        for gridtype in self.grids:
            # Check if any variable in the grid contains 'bnds'
            if not gridtype.dims:
                removed.append(gridtype)
                self.loggy.info('Removing the grid defined by %s with with no spatial dimensions',
                                gridtype.dims)
            # elif all('bnds' in variable for variable in gridtype.variables):
            #    removed.append(gridtype)  # Add to removed list
            #    self.loggy.info('Removing the grid defined by %s with variables containing "bnds"',
            #                     gridtype.dims)
            # elif all('bounds' in variable for variable in gridtype.variables):
            #    removed.append(gridtype)  # Add to removed list
            #    self.loggy.info('Removing the grid defined by %s with variables containing "bounds"',
            #                    gridtype.dims)

        for remove in removed:
            self.grids.remove(remove)

    def _identify_variable(self, gridtype, var_data, var_name=None):
        """Helper function to process individual variables.

        Args:
            gridtype (GridType): The Gridtype object of originally inspected from the data
            var_data (xr.DataArray): An xarray DataArray containing variable data.
            var_name (str, optional): The name of the variable. If None, uses the name from var_data.

        Updates:
            self.variables: Updates the variables dictionary with the variable's coordinates.
        """
        datagridtype = GridType(var_data.dims, extra_dims=self.extra_dims)

        if datagridtype == gridtype:
            gridtype.variables[var_name or var_data.name] = {
                'coords': list(var_data.coords),
            }

    def identify_variables(self, gridtype):
        """
        Identify variables in the provided data that match the defined dimensions.

        Args:
            gridtype (GridType): The Gridtype object of originally inspected from the data

        Raises:
            TypeError: If the input data is neither an xarray Dataset nor DataArray.

        Updates:
            self.variables: Updates the variables dictionary with identified variables and their coordinates.
            self.bounds: Updates the bounds list with identified bounds variables from the dataset.
        """

        if isinstance(self.data, xr.Dataset):
            for var in self.data.data_vars:
                self._identify_variable(gridtype, self.data[var], var)
            gridtype.bounds = self._identify_spatial_bounds(self.data)
            gridtype.variables = {
                key: value
                for key, value in gridtype.variables.items()
                if key not in set(gridtype.bounds)
            }

        elif isinstance(self.data, xr.DataArray):
            self._identify_variable(gridtype, self.data)

    def _is_bounds(self, var):
        """
        Check if a variable is a bounds variable.
        """
        return var.endswith('_bnds') or var.endswith('_bounds') or var == 'vertices' and 'time' not in var

    def _identify_spatial_bounds(self, data):
        """
        Identify bounds variables in the dataset by checking variable names.

        Args:
            data (xr.Dataset): An xarray dataset containing data variables.

        Returns:
            list: A list of bounds variable names identified in the dataset.
        """
        bounds_variables = []

        for var in data.data_vars:
            if self._is_bounds(var):
                bounds_variables.append(var)

        return bounds_variables

    def _identify_grid_format(self, data):
        """
        Identify the grid format based on the provided data.

        Args:
            data (xr.Dataset or xr.DataArray): The input dataset.

        Returns:
            str: The identified grid format.
        """
        for gridtype in self.grids:
            variables = list(gridtype.variables.keys())
            if variables:
                if isinstance(data, xr.Dataset):
                    gridtype.kind = self.detect_grid(data[variables])
                elif isinstance(data, xr.DataArray):
                    gridtype.kind = self.detect_grid(data)

    @staticmethod
    def get_gridtype_attr(gridtypes, attr):
        """Helper compact tool to extra gridtypes information"""
        out = []
        for gridtype in gridtypes:
            value = getattr(gridtype, attr, None)
            if isinstance(value, (list, tuple)):
                out.extend(value)
            elif isinstance(value, dict):
                out.extend(value.keys())
            elif isinstance(value, str):
                out.append(value)

        return list(dict.fromkeys(out))

    def detect_grid(self, data, lat='lat', lon='lon'):
        """
        Detect the grid type based on the structure of the data.

        Args:
            data (xr.Dataset or xr.DataArray): The input dataset.
            lat (str): The name of the latitude coordinate.
            lon (str): The name of the longitude coordinate.

        Returns:
            str: The identified grid type.
        """

        lon = find_coord(data, set(LON_COORDS + [lon]))
        lat = find_coord(data, set(LAT_COORDS + [lat]))

        if self.is_healpix_from_attribute(data):
            return "HEALPix"

        if not lat or not lon:
            return "Unknown"

        # 2D coord-dim dependency
        if data[lat].ndim == 2 and data[lon].ndim == 2:
            return "Curvilinear"

        # 1D coord-dim dependency
        if data[lat].ndim == 1 and data[lon].ndim == 1:

            # Regular: latitude and longitude depend on different coordinates
            if data[lat].dims != data[lon].dims:

                lat_diff = np.diff(data[lat].values)
                lon_diff = np.diff(data[lon].values)

                # Regular: latitude and longitude equidistant
                if np.allclose(lat_diff, lat_diff[0]) and np.allclose(lon_diff, lon_diff[0]):
                    return "Regular"

                # Gaussian: longitude equidistant, latitude not
                if not np.allclose(lat_diff, lat_diff[0]) and np.allclose(lon_diff, lon_diff[0]):
                    return "GaussianRegular"

                return "UndefinedRegular"

            pix = data[lat].size
            if pix % 12 == 0 and (pix // 12).bit_length() - 1 == np.log2(pix // 12):
                return "HEALPix"

            # Guess gaussian reduced: increasing number of latitudes from -90 to 0
            lat_values = data[lat].where(data[lat] < 0).values
            lat_values = lat_values[~np.isnan(lat_values)]
            _, counts = np.unique(lat_values, return_counts=True)
            gaussian_reduced = np.all(np.diff(counts) > 0)
            if gaussian_reduced:
                return "GaussianReduced"

            # None of the above cases
            return "Unstructured"

        return "Unknown"

    @staticmethod
    def is_healpix_from_attribute(data):
        """
        Determine if the given xarray Dataset or DataArray uses a HEALPix grid.

        Returns:
            bool: True if HEALPix grid detected, False otherwise.
        """

        # Attribute-based checks
        if isinstance(data, xr.Dataset):
            if "healpix" in data.variables:
                return True
            for var in data.data_vars:
                if data[var].attrs.get('grid_mapping') == 'healpix':
                    return True
        elif isinstance(data, xr.DataArray):
            if data.attrs.get('grid_mapping') == 'healpix':
                return True

        return False



"""Weight calculation utilities"""

import dask
import sparse


def compute_weights_matrix3d(weights, mask_dim='lev'):
    """
    Convert the weights from CDO to a list of numpy arrays
    """

    # CDO 2.2.0 fix
    links_dim = "numLinks" if "numLinks" in weights.dims else "num_links"

    # precompute link lengths
    link_length = weights["link_length"].values

    sparse_weights = []
    for i, nl in enumerate(link_length):
        w = weights.isel({mask_dim: i, links_dim: slice(0, nl)})
        sparse_weights.append(compute_weights_matrix(w))

    return sparse_weights

def compute_weights_matrix(weights):
    """
    Convert the weights from CDO to a numpy array
    """

    # CDO style weights
    src_address = weights.src_address - 1
    dst_address = weights.dst_address - 1
    remap_matrix = weights.remap_matrix[:, 0]
    w_shape = (weights.sizes["src_grid_size"], weights.sizes["dst_grid_size"])

    # Create a sparse array from the weights
    sparse_weights_delayed = dask.delayed(sparse.COO)(
        [src_address.data, dst_address.data], remap_matrix.data, shape=w_shape
    )
    sparse_weights = dask.array.from_delayed(
        sparse_weights_delayed, shape=w_shape, dtype=remap_matrix.dtype
    )

    return sparse_weights


def mask_tensordot(src_mask, weights_matrix):
    """Apply tensor dot product to source mask to return destination mask"""

    target_mask = dask.array.tensordot(src_mask, weights_matrix, axes=1)
    target_mask = dask.array.where(target_mask < 0.5, 0, 1)
    return target_mask


def mask_weights(weights, weights_matrix, mask_dim=None):
    """This functions precompute the mask for the target interpolation
    Takes as input the weights from CDO and the precomputed weights matrix
    Return the target mask: handle the 3d case.
    
    Claude-based solution with full lazy evaluation and no in-place modification.
    
    Args:
        weights: xarray Dataset containing the weights and source mask.
        weights_matrix: Precomputed weights matrix (sparse arrays).
        mask_dim: Name of the mask dimension if 3D regridding is used.
    """

    src_mask = weights.src_grid_imask
    if mask_dim is not None:
        # Create list of masked arrays lazily
        masked_levels = [
            mask_tensordot(src_mask.isel({mask_dim: i}).data, weights_matrix[i])
            for i in range(weights.sizes[mask_dim])
        ]
        # Stack all levels at once (lazy operation)
        dst_mask = dask.array.stack(masked_levels, axis=0)
        # Create new DataArray without in-place modification
        weights = weights.assign(dst_grid_imask=([mask_dim] + list(weights['dst_grid_imask'].dims[1:]), dst_mask))
    else:
        mask = src_mask.data
        dst_mask = mask_tensordot(mask, weights_matrix)
        weights = weights.assign(dst_grid_imask=(weights['dst_grid_imask'].dims, dst_mask))

    return weights


# def mask_weights(weights, weights_matrix, mask_dim=None):
#     """This functions precompute the mask for the target interpolation
#     Takes as input the weights from CDO and the precomputed weights matrix
#     Return the target mask: handle the 3d case"""

#     src_mask = weights.src_grid_imask
#     if mask_dim is not None:
#         for i in range(weights.sizes[mask_dim]):
#             mask = src_mask.isel({mask_dim: i}).data
#             weights['dst_grid_imask'].isel({mask_dim: i}).data[:] = mask_tensordot(mask, weights_matrix[i])
#     else:
#         mask = src_mask.data
#         weights['dst_grid_imask'].data = mask_tensordot(mask, weights_matrix)

#     return weights

def check_mask(weights, mask_dim=None):
    """
    Check if the target mask has all values equal to 1.
    Returns a boolean indicating whether the mask is full.
    Handles the 3D case efficiently.
    """
    wdst = weights['dst_grid_imask']

    if mask_dim is not None:
        # Reduce along all dimensions except the vertical dimension
        dims_to_reduce = [dim for dim in wdst.dims if dim != mask_dim]
        mask = ~(wdst == 1).all(dim=dims_to_reduce)
    else:
        # Check if all values are 1 across all dimensions
        mask = ~(wdst == 1).all()

    # Compute the result if it's a dask array
    return mask.compute()
